##########################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: Composite.py,v 1.2 2009-06-09 14:31:42 moscicki Exp $
#
# Copyright (C) 2003-2007 The Ganga Project
#
# This file is part of GangaCore.
#
# Ganga is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Ganga is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
##########################################################################

import sys
from GangaCore.GPIDev.Adapters.IMonitoringService import IMonitoringService

class CompositeMonitoringService(IMonitoringService):

    """ IMonitoringService container:
    Wrapper object containing a list of IMonitoringService(s) inside and delegating
    the interface methods to each of them. (composite design pattern) 
    This object is used automatically to transparently wrap the list of monitoring services set in the configuration    
    """

    def __init__(self, monClasses, jobInfos, configInfos):
        """Create a new composite monitoring service based on the lists of
        monitoring classes, jobs and configs (all the same length).

        If this is called in the Ganga client, i.e. from Ganga/GPIDev/MonitoringServices,
        then jobInfos is a list of Job (all the same), configInfos is a list of
        Config (specific to each monitoring class).

        If this is called on the worker node, i.e. from the text generated by
        getWrapperScriptConstructorText(), the jobInfos are dictionaries (specific
        to each monitoring class) and configInfos are dictionaries of effective
        config options (specific to each monitoring class).
        """

        if not (len(monClasses) == len(jobInfos) == len(configInfos)):
            raise Exception(
                "cannot create monitoring object, list of monitoring classes, jobs and configs are not the same length.")

        IMonitoringService.__init__(self, jobInfos, configInfos)

        # init the logger
        try:
            import GangaCore.Utility.logging
            self.logger = GangaCore.Utility.logging.getLogger()
        except ImportError:
            # on the worker node we don't have access to Ganga logging facilities
            # so we simple print out the log message
            self.logger = None

        # init the monitoring services
        self.monMonServices = []
        for i in range(len(monClasses)):
            try:
                monClass = monClasses[i]
                # allow for existing monitoring classes which do not take
                # config_info in constructor
                if configInfos[i] is None:
                    monService = monClass(jobInfos[i])
                else:
                    monService = monClass(jobInfos[i], configInfos[i])
                self.monMonServices.append(monService)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("Failed to init %s monitoring service...discarding it" % str(monClass))
                from GangaCore.Utility.logging import log_user_exception
                log_user_exception(self.logger)

    def start(self, **opts):
        """Application is about to start on the worker node.
        Called by: job wrapper.
        """
        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.start(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
               self.logger.warning("%s monitoring service failed to *start*: %s" % (monClass, e))

        return ret

    def progress(self, **opts):
        """Application execution is in progress (called periodically, several times a second).
        Called by: job wrapper. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.progress(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed to *progress*: %s" % (monClass, e))

        return ret

    def stop(self, exitcode, **opts):
        """Application execution finished.
        Called by: job wrapper. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.stop(exitcode, **opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed to *stop*: %s" % (monClass, e))
        return ret

    def prepare(self, **opts):
        """Preparation of a job.
        Called by: ganga client. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.prepare(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed in job *prepare*" % monClass)
        return ret

    def submitting(self, **opts):
        """Submission of a job.
        Called by: ganga client. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.submitting(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed in job *submitting*" % monClass)
        return ret

    def submit(self, **opts):
        """Submission of a job.
        Called by: ganga client. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.submit(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed in job *submit*" % monClass)
                from GangaCore.Utility.logging import log_user_exception
                log_user_exception(self.logger)
        return ret

    def complete(self, **opts):
        """Completion of a job.
        Called by: ganga client. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.complete(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed in job *complete*" % monClass)
        return ret

    def fail(self, **opts):
        """Failure of a job.
        Called by: ganga client. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.fail(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed in job *fail*" % monClass)
        return ret

    def kill(self, **opts):
        """Killing of a job.
        Called by: ganga client. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.kill(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed in job *kill*" % monClass)
        return ret

    def rollback(self, **opts):
        """Rollback of a job to new state (caused by error during submission).
        Called by: ganga client. """

        ret = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                ret[monClass] = monService.rollback(**opts)
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed in job *rollback*" % monClass)
        return ret

    def getJobInfo(self):
        """ Return a static info object which static information about the job
        at submission time. Called by: ganga client.

        The info object is passed to the contructor. Info
        object may only contain the standard python types (such as lists,
        dictionaries, int, strings). 
        Implementation details:
         return the job info objects as a map for each compound Monitoring Service
         @see getWrapperScriptConstructorText() method
        """

        infos = {}
        for monService in self.monMonServices:
            try:
                monClass = str(monService.__class__)
                infos[monClass] = monService.getJobInfo()
            except Exception as e:
                # discard errors in initialization of monitoring services
                self.logger.warning("%s monitoring service failed in *getJobInfo*: %s" % (monClass, e))
        return infos




import re
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Schema import *
from GangaCore.Utility.Config import getConfig

class NA62LCGRequirements(GangaObject):
   '''Helper class for NA62 specific LCG requirements.

   See also: JDL Attributes Specification at http://cern.ch/glite/documentation
   '''

   _schema = Schema(Version(1,2), {
      'memory'          : SimpleItem(defvalue=0,doc='Mininum available memory (MB)'),
      'cputime'         : SimpleItem(defvalue=0,doc='Minimum available CPU time (min)'),
      'walltime'        : SimpleItem(defvalue=0,doc='Mimimum available total time (min)'),
      'allowedSites'    : SimpleItem(defvalue=[], typelist=['str'],sequence=1, doc='list of allowed sites'),
      'excludedSites'   : SimpleItem(defvalue=[], typelist=['str'],sequence=1,doc='list of excluded sites'),
      'other'           : SimpleItem(defvalue=[],typelist=['str'],sequence=1,doc='Other Requirements')
   })

   _category = 'LCGRequirements'
   _name = 'NA62LCGRequirements'

   def __init__(self):      
      super(NA62LCGRequirements,self).__init__()

   def merge(self,other):
      '''Merge requirements objects'''
      
      if not other: return self
      
      merged = NA62LCGRequirements()
      for name in [ 'memory', 'cputime', 'walltime', 'allowedSites', 'excludedSites', 'other' ]:

         attr = ''

         try:
             attr = getattr(other,name)
         except KeyError as e:
             pass

         if not attr: attr = getattr(self,name)
         setattr(merged,name,attr)
         
      return merged
   
   def convert(self):
      '''Convert the condition in a JDL specification'''
     
      import re

      ##requirements = [ 'Member("%s",other.GlueHostApplicationSoftwareRunTimeEnvironment)' % sw for sw in self.software ]
      requirements = []
      if self.memory:         requirements += [ 'other.GlueHostMainMemoryRAMSize >= %s' % str(self.memory) ]
      if self.cputime:        requirements += [ '(other.GlueCEPolicyMaxCPUTime >= %s || other.GlueCEPolicyMaxCPUTime == 0)' % str(self.cputime) ]
      if self.walltime:       requirements += [ '(other.GlueCEPolicyMaxWallClockTime >= %s || other.GlueCEPolicyMaxWallClockTime == 0)' % str(self.walltime) ]
      requirements += self.other

      # map site names to CEs
      # should be done with a DB call or something...
      site_map  = { 'LIV': 'liv.ac.uk',
                    'BIR': 'bham.ac.uk',
                    'GLA': 'scotgrid.ac.uk',
                    'IC': 'ic.ac.uk',
                    'RAL': 'rl.ac.uk',
                    'UCL': 'ucl.ac.uk' }

      allowed_ces = []
      excluded_ces = []
      for s in self.allowedSites:
         allowed_ces.append( site_map[s] )

      for s in self.excludedSites:         
         excluded_ces.append( site_map[s] )

      ## composing the requirements given the list of allowed_ces and excluded_ces
      if allowed_ces:
          requirements += [ '( %s )' % ' || '.join([ 'RegExp("%s",other.GlueCEUniqueID)' % ce for ce in allowed_ces])]
          
      if excluded_ces:
          #requirements += [ '(!RegExp("%s",other.GlueCEUniqueID))' % ce for ce in excluded_ces ]
          requirements += [ '( %s )' % ' && '.join([ '(!RegExp("%s",other.GlueCEUniqueID))' % ce for ce in excluded_ces])]

      return requirements



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ElapsedTimeProfiler.py,v 1.1 2008-07-17 16:40:57 moscicki Exp $
###############################################################################
#
# LCG backend profiler
#
# ATLAS/ARDA
#
# Date:   November 2007
import time
from GangaCore.Utility.logging import getLogger


class ElapsedTimeProfiler(object):

    '''Elapsed time profiler'''

    def __init__(self, logger=None):

        if not logger:
            logger = getLogger(name='GangaCore.Lib.LCG.ElapsedTimeProfiler')

        self.logger = logger

    def start(self):
        self.beg = time.time()

    def check(self, message):
        etime = time.time() - self.beg
        self.logger.debug('%s: %f sec.' % (message, etime))

    def checkAndStart(self, message):
        self.check(message)
        self.start()



from GangaCore.GPIDev.Adapters.IApplication import IApplication
from GangaCore.GPIDev.Schema import *

import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger()

class ProdTrans(IApplication):
    """Ganga application for production queues."""

    _schema = Schema(Version(1, 0), {
            'atlas_release': SimpleItem(defvalue='',
                                        doc='ATLAS Software Release'),
            'atlas_cmtconfig': SimpleItem(defvalue='',
                                          doc='CMTCONFIG environment variable'),
            'output_files': SimpleItem(defvalue=[],
                                       typelist=['str'],
                                       sequence=1,
                                       doc='Logical File Names for output'),
            'home_package': SimpleItem(defvalue='',
                                       doc='Home package'),
            'transformation': SimpleItem(defvalue='',
                                         doc='Transformation'),
            'job_parameters': SimpleItem(defvalue='',
                                         doc='Addtional parameters for the job'),
            'prod_source_label': SimpleItem(defvalue='',
                                            doc='Production Source Label'),
            'dbrelease': SimpleItem(defvalue='LATEST',
                                    doc='ATLAS DB Release. Use LATEST for most recent'),
            'dbrelease_dataset': SimpleItem(defvalue='',
                                            doc='Dataset name for the DB'),
            'priority': SimpleItem(defvalue=1000,
                                   doc='Initialial priority for the Job'),
            'randomize_lfns': SimpleItem(defvalue=False,
                                         doc='Generate output file names with random suffix'),
            'max_events': SimpleItem(defvalue=0,
                                     docs='Max events'),
            'output_type': SimpleItem(defvalue=['NTUP_TOP'], typelist=['str'],sequence=1, strict_sequence=0,
                                      doc='Type of the output file'),
            'input_type': SimpleItem(defvalue='',
                                     doc='Type of the input file'),
            'is_prepared': SharedItem(defvalue=None, strict_sequence=0, visitable=1, 
                copyable=1, typelist=['type(None)', 'bool', 'str'], protected=0,
                doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
            'core_count': SimpleItem(defvalue=0,
                                     docs='Number of cores to run.'),
            })
    _category = 'applications'
    _name = 'ProdTrans'
    _exportmethods = ['postprocess', 'unprepare']

    def __init__(self):
        super(ProdTrans, self).__init__()
        logger.debug('Initializing ProdTrans object')

    def master_configure(self):
        """Configure the general aspect of the application."""
        logger.debug('ProdTrans master_configure called')
        return (0, None)

    def configure(self, master_appconfig):
        """Configure the specific aspect of the application."""
        logger.debug('ProdTrans configure called')
        return (None, None)

    def postprocess(self):
        """Determine the outputdata and outputsandbox locations."""
        pass



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: GListApp.py,v 1.2 2008-10-22 11:59:19 wreece Exp $
##########################################################################
from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema.Schema import ComponentItem, Schema, SimpleItem, Version
from GangaCore.GPIDev.Lib.File import File, ShareDir


class GListApp(IPrepareApp):

    """Test File object with well known equality properties -i.e. Does not reply on proxy!"""
# summary_print
    _category = 'applications'
    _exportedmethods = ['configure']
    _name = 'GListApp'
    _schema = Schema(Version(1, 0), {
        'bound_print_comp': ComponentItem('files', defvalue=[], sequence=1, summary_print='_print_summary_bound_comp', typelist=['str', 'GangaCore.test.GPI.GangaList.TFile.TFile']),
        'bound_print_simple': SimpleItem(defvalue=[], sequence=1, summary_print='_print_summary_bound_simple'),
        'no_summary': SimpleItem(defvalue=[], sequence=1, summary_sequence_maxlen=-1, typelist=['str']),
        'seq': SimpleItem(defvalue=[], sequence=1, typelist=['int']),
        'gList': SimpleItem(defvalue=[], sequence=1, typelist=['str']),
        'gListComp': ComponentItem('files', defvalue=[], sequence=1),
        'simple_print': SimpleItem(defvalue='', summary_print='_print_summary_simple_print'),
        'comp_print': ComponentItem('backends', defvalue=None, summary_print='_print_summary_comp_print'),
        'is_prepared': SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, hidden=0, typelist=['type(None)', 'bool', ShareDir], protected=0, comparable=1, doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
    })

    def configure(self, master_appconfig):
        return (None, None)

    def _print_summary_bound_comp(self, value, verbosity_level):
        return '_print_summary_bound_comp'

    def _print_summary_bound_simple(self, value, verbosity_level):
        return '_print_summary_bound_simple'

    def _print_summary_simple_print(self, value, verbosity_level):
        return '_print_summary_simple_print'

    def _print_summary_comp_print(self, value, verbosity_level):
        return '_print_summary_comp_print'


class Handler(IRuntimeHandler):

    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        return 0
allHandlers.add('GListApp', 'TestSubmitter', Handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ============================================================================
## @file
#  Application handler for Ostap
#
# The user specifies:
#  - script file(s) which contain Ostap scripts,
#  - set of command line arguments
#  
# At least one 'script' or non-empty list of commands is required 
#
# The application executes the following line:
# @code   
# % ostap {scripts} {arguments} --import {imports} --no-color --batch
# @endcode    
# e.g.
# @code 
# % bender script.py  --with-context -p5 --no-color --no-canvas --batch
# @endcode
#
# 
# @code
# my_script  = "~/cmtuser/tests/my_batch.py"
# my_app     = Ostap (
#    version      = 'v28r3'       ,
#    scripts      =   my_script   ,
#    arguments    = [
#      ] , 
#    commands     = [ 'print dir()']
#      ) 
# @endcode
#
# @author Vladimir ROMANOVSKY  Vladimir.Romanovskiy@cern.ch
# @author Vanya BELYAEV        Ivan.Belyaev@itep.ru
# @fate 2016-03-16
#
# Version           $Revision:$
# Last Modification $Date:$
#                by $Author:$
# =============================================================================
"""The application handler for Ostap

The user specifies:
- script file(s) which contain Ostap scripts,
- set of command line arguments

At least one 'script' or non-empty list of commands is required 

The application executes the following line:

% ostap {scripts} {arguments} --import {imports} --no-color --batch

e.g.
% ostap script.py  --with-context -p5 --no-color --no-canvas --batch

>>> my_script  = '~/cmtuser/tests/my_batch.py'
>>> my_app     = Ostap (
...    version      = 'v28r3'       ,
...    scripts      =   my_script   ,
...    arguments    = [
...      '--with-context'           , ## execute the script files ``with-context''
...      '--no-canvas'              , ## do not create canvas 
...      ] , 
...    commands     = [ 'print dir()']
... ) 
"""
# =============================================================================
__date__    = '2016-03-16'
__version__ = '$Revision:$'
__author__  = 'Vladimir ROMANOVSKY, Vanya BELYAEV'
# =============================================================================
import os
from   os.path import split, join
from   GangaCore.GPIDev.Schema.Schema             import FileItem, SimpleItem
from   GangaCore.GPIDev.Lib.File                  import File
from   GangaCore.Utility.util                     import unique
from   GangaCore.Core.exceptions                  import ApplicationConfigurationError
from   GangaCore.GPIDev.Lib.File.FileBuffer       import FileBuffer
from   GangaGaudi.Lib.Applications.GaudiBase  import GaudiBase
from   GangaGaudi.Lib.Applications.GaudiUtils import fillPackedSandbox, gzipFile
from   GangaCore.Utility.files                    import expandfilename, fullpath
from   GangaCore.Utility.Config                   import getConfig
from   .AppsBaseUtils                          import guess_version
#
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

# Added for XML PostProcessing
from GangaLHCb.Lib.Applications import XMLPostProcessor

import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger()

# =============================================================================
## the actual wrapper script to execute 
# =============================================================================
layout = """#!/usr/bin/env python
# -*- coding: utf-8 -*-
# =============================================================================
from distutils.spawn import find_executable
ostap_script = find_executable('ostap')
import sys
# =============================================================================
## redefine arguments 
# =============================================================================
sys.argv += {scripts}        
sys.argv += {arguments}
sys.argv += [ '--no-color'  , '--batch' ]
sys.argv += [ '--command' ] + {command} 
##
# =============================================================================
## execute bender script 
# =============================================================================
execfile(ostap_script)
# =============================================================================
# The END
# =============================================================================
"""
# =============================================================================
# @class Ostap
# Application handler for Ostap
#
# The user specifies:
#  - script file(s) which contain BenderScript scripts,
#  - configuration/Configurables files, to be used for 'importOptions'
#  - set of command-line arguments
# 
# At least one 'script' or 'import' file is required.
#
# The application executes the following line:
# @code    
# % ostap {scripts} {arguments} --no-color --no-castor --batch
# @endcodce    
# e.g.
# @code 
# % ostap script1.py  --no-color --batch
# @endcode 
#
# @author Vladimir ROMANOVSKY  Vladimir.Romanovskiy@cern.ch
# @author Vanya BELYAEV        Ivan.Belyaev@itep.ru 
class Ostap(GaudiBase):    
    """The application handler for Ostap
    
    The user specifies:
    - script file(s) which contain Ostap scripts,
    - set of command line arguments
    
    At least one 'script' or non-empty list of commands is required 
    
    The application executes the following line:
    
    % ostap {scripts} {arguments} --import {imports} --no-color --batch
    
    e.g.
    % ostap script.py -p5 --no-color --no-canvas --batch
    
    >>> my_script  = '~/cmtuser/tests/my_batch.py'
    >>> my_app     = Ostap (
    ...    project      = 'Analysis'    , ## 
    ...    version      = 'v16r1'       , ## version of Analysis
    ...    scripts      =   my_script   ,
    ...    arguments    = [
    ...      '--no-canvas'              , ## do not create canvas 
    ...      ] , 
    ...    commands     = [ 'print dir()']
    ... ) 
    """
    
    _name           = 'Ostap'
    _category       = 'applications'
    _exportmethods  = GaudiBase._exportmethods[:]
    _exportmethods += ['prepare', 'unprepare']
    
    _schema = GaudiBase._schema.inherit_copy()

    _schema.datadict['package'] = SimpleItem(
        defvalue = None,
        typelist = ['str', 'type(None)'],
        doc      = """The package the application belongs to (e.g. 'Sim', 'Phys')
        """
        )
    _schema.datadict['masterpackage'] = SimpleItem (
        defvalue = None,
        typelist = [ 'str', 'type(None)' ],
        doc      = """The package where your top level requirements file is read from.
        Can be written either as a path 'Tutorial/Analysis/v6r0' or in traditional notation 
        'Analysis v6r0 Tutorial'
        """
        )
    
    _schema.datadict['setupProjectOptions'] = SimpleItem(
        defvalue = ''     ,
        typelist = [ 'str', 'type(None)'],
        doc      = """Extra options to be passed onto the SetupProject command
        used for configuring the environment. As an example 
        setting it to '--dev' will give access to the DEV area. 
        For full documentation of the available options see 
        https://twiki.cern.ch/twiki/bin/view/LHCb/SetupProject
        """
        )
    
    _schema.datadict['scripts'] = FileItem(
        preparable      = 1      ,
        sequence        = 1      ,
        strict_sequence = 0      ,
        defvalue        = []     ,
        doc             = """The names of the script files to execute.
        A copy will be made at submission time
        """
        )
    
    _schema.datadict['commands'] = SimpleItem(
        defvalue = []      ,
        typelist = ['str'] ,
        sequence =  1      ,
        doc      = """The commands to be executed,
        e.g. [ 'run(10)' , 'print ls()' , 'print dir()' ]
        """
        )
    
    _schema.datadict['arguments'] = SimpleItem(
        defvalue = []      ,
        typelist = ['str'] ,
        sequence =  1      ,
        doc      = """List of command-line arguments for bender script,
        e.g. ['-w','-p5'], etc.
        For python scripts and configuration/Configurable files for 'importOptions'
        it is much better to use the separate options 'scripts' and 'imports'
        Following arguments will be appended automatically:  --no-color, --no-castor and --batch
        """
        )
    
    _schema.version.major += 2
    _schema.version.minor += 0
    
    def _get_default_version(self, gaudi_app):
        return guess_version(self, gaudi_app)

    def _auto__init__(self):
        if not self.appname : self.appname = 'Bender'  # default
        self._init()

    def _getshell(self):

        from . import EnvironFunctions
        return EnvironFunctions._getshell(self)

    def prepare(self, force=False):

        super(Ostap, self).prepare(force)
        self._check_inputs()

        
        share_dir = os.path.join (
            expandfilename ( getConfig('Configuration')['gangadir'] ) ,
            'shared'                            ,
            getConfig('Configuration')['user']  ,
            self.is_prepared.name               )
        
        input_sandbox_tar = os.path.join ( share_dir , 'inputsandbox',
                                           '_input_sandbox_%s.tar' % self.is_prepared.name ) 
        input_sandbox_tgz = os.path.join ( share_dir , 'inputsandbox',
                                           '_input_sandbox_%s.tgz' % self.is_prepared.name ) 
        
        fillPackedSandbox ( self.scripts      , input_sandbox_tar        ) 
        gzipFile          ( input_sandbox_tar , input_sandbox_tgz , True )
        
        # add the newly created shared directory into the metadata system if
        # the app is associated with a persisted object
        self.checkPreparedHasParent(self)
        self.post_prepare()
        logger.debug("Finished Preparing Application in %s" % share_dir)

    def master_configure(self):
        return (None, StandardJobConfig())

    def configure(self, master_appconfig):
        
        ## strip leading and trailing blanks from arguments 
        self.arguments = [ a.strip() for a in self.arguments ]

        ## strip leading and trailing blanks from the command 
        self.commands  = [ a.strip() for a in self.commands  ]
        
        ## the script layout
        the_script    = layout.format (
            scripts   = [ os.path.join ( f.subdir , os.path.basename ( f.name ) ) for f in self.scripts ] , 
            arguments = self.arguments  ,
            command   = self.commands    
            )

        # add summary.xml
        outputsandbox_temp  = XMLPostProcessor._XMLJobFiles()
        outputsandbox_temp += unique(self.getJobObject().outputsandbox)
        outputsandbox       = unique(outputsandbox_temp)
        
        input_files  = []
        input_files += [ FileBuffer('gaudipython-wrapper.py', the_script ) ]
        logger.debug("Returning StandardJobConfig")
        return (None, StandardJobConfig(inputbox=input_files,
                                        outputbox=outputsandbox))
    
    def _check_inputs(self):
        """Checks the validity of user's entries for Ostap schema"""
        
        if not self.scripts and not self.commands and not self.arguments : 
            raise ApplicationConfigurationError("Application scripts are not defined")
        
        if isinstance ( self.scripts , str ) : self.scripts = [ File ( self.scripts ) ]        
        for f in self.scripts : f.name = fullpath ( f.name )

    def postprocess(self):
        XMLPostProcessor.postprocess(self, logger)

# =============================================================================
# Associate the correct run-time handlers to GaudiPython for various backends.
# =============================================================================

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaLHCb.Lib.RTHandlers.LHCbGaudiRunTimeHandler import LHCbGaudiRunTimeHandler
from GangaLHCb.Lib.RTHandlers.LHCbGaudiDiracRunTimeHandler import LHCbGaudiDiracRunTimeHandler

for backend in ['LSF', 'Interactive', 'PBS', 'SGE', 'Local', 'Condor', 'Remote']:
    allHandlers.add('Ostap', backend, LHCbGaudiRunTimeHandler)
allHandlers.add('Ostap', 'Dirac', LHCbGaudiDiracRunTimeHandler)


# =============================================================================
if '__main__' == __name__ :

    print(80*'*')  
    print(__doc__) 
    print(' Author  : %s ' %  __author__)   
    print(' Version : %s ' %  __version__)  
    print(' Date    : %s ' %  __date__)     
    print(80*'*')  

# =============================================================================
# The END 
# =============================================================================



# NOTICE:
# this module is used by config system to configure logging therefore the logger
# requires special handling
# this module must be importable without any side-effect, especially without
# creation of loggers by surrounding packages...

from inspect import isclass

found_types = {}
found_values = {}

def dummy_func( _input ):
    return _input

def safer_eval( _input ):
    try:
        from GangaCore.GPIDev.Base.Proxy import getRuntimeGPIObject
        temp_output = getRuntimeGPIObject(_input, True)
        if temp_output is None:
            if len(_input) > 0:
                try:
                    if not isinstance(_input, str):
                        _input = str(_input)
                    _output = eval(_input)
                except:
                    _output = str(_input)
            else:
                _output = None
        elif isclass(temp_output):
            _output = stripProxy(temp_output)
        else:
            _output = temp_output
    except ImportError:
        if len(_input) > 0:
            try:
                if not isinstance(_input, str):
                    _input = str(_input)
                _output = eval(_input)
            except:
                _output = str(_input)
        else:
            _output = None

    return _output

def _valueTypeAllowed(val, valTypeList, logger=None):
    for _t in valTypeList:

        ## Return None when None
        if _t is None:
            if val is None:
                return True

        if isinstance(_t, str):
            global found_types

            if _t not in found_types:
                temp = safer_eval(_t)
                if not isinstance(temp, type(str)):
                    temp = type(temp)
                found_types[_t] = temp
            _type = found_types[_t]

            if _type == str:
                GangaSplit = _t.split('.')
                if len(GangaSplit) > 1:
                    from GangaCore.Utility.util import importName

                    imported_Class = importName( '.'.join(GangaSplit[:-1]), GangaSplit[-1])

                    if imported_Class is not None:
                        _type = imported_Class

        else:
            _type = _t
        
        if isinstance(val, str):
            global found_values

            if val not in found_values:
                found_values[val] = safer_eval(val)
            _val = found_values[val]
        else:
            _val = val

        try:
            from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
            knownLists = (list, tuple, GangaList)
        except Exception as err:
            knownLists = (list, tuple)

        # Deal with the case where type is None.
        if _type is None:
            if _val is None:
                return True
            else:
                continue

        try:
            ## Type Checking ""Should"" be proxy agnoistic but this may have problems loading before certain classes
            from GangaCore.GPIDev.Base.Proxy import stripProxy
        except ImportError:
            stripProxy = dummy_func

        try:
            from GangaCore.GPIDev.Base.Objects import GangaObject
        except:
            GangaObject = None

        raw_type = stripProxy(_type)
        raw_val = stripProxy(val)

        if isinstance(raw_type, GangaObject):
            return isinstance(raw_val, raw_type)

        try:
            if raw_type in knownLists:
                if isinstance(raw_val, knownLists):
                    return True
            elif isinstance(raw_val, raw_type):
                return True
        except:
            continue

    return False




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Notebook.py,v 1.1 $
##########################################################################

from os import path
import inspect
import uuid

from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaCore.GPIDev.Lib.File import FileUtils
from GangaCore.GPIDev.Lib.File import FileBuffer
from GangaCore.GPIDev.Lib.File import ShareDir
from GangaCore.Core.exceptions import ApplicationPrepareError
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Base.Proxy import getName


logger = getLogger()

class Notebook(IPrepareApp):

    """Notebook application -- execute Jupyter notebooks.

    All cells in the notebooks given as inputfiles will be evaluated 
    and the results returned in the same notebooks.

    A simple example is

    app = Notebook()
    infiles = [LocalFile('/abc/test.ipynb')]
    outfiles = [LocalFile('test.ipynb')]
    j = Job(application=app, inputfiles=files, backend=Local())
    j.submit()

    The input can come from any GangaFile type supported and the same
    is the case for the output.

    All inputfiles matching the regular expressions (default all
    files ending in .ipynb) given are executed. Other files will
    simply be unpacked and available.

    """
    _schema = Schema(Version(1, 0), {
        'version': SimpleItem(preparable=1, defvalue=None, typelist=[None, int], doc="Version of the notebook. If None, it will be assumed that it is the latest one."),
        'timeout': SimpleItem(preparable=1, defvalue=None, typelist=[None, int], doc="Timeout in seconds for executing a notebook. If None, the default value will be taken."),
        'kernel': SimpleItem(preparable=1, defvalue='python2', doc="The kernel to use for the notebook execution. Depending on configuration, python3, Root and R might be available."),
        'regexp': SimpleItem(preparable=1, defvalue=[r'.+\.ipynb$'], typelist=["str"], sequence=1, strict_sequence=0, doc="Regular expression for the inputfiles to match for executing."),
       'is_prepared': SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, hidden=0, typelist=[None, ShareDir], protected=0, comparable=1, doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        'hash': SimpleItem(defvalue=None, typelist=[None, str], hidden=0, doc='MD5 hash of the string representation of applications preparable attributes')
    })
    _category = 'applications'
    _name = 'Notebook'
    _exportmethods = ['prepare', 'unprepare']

    def __init__(self):
        super(Notebook, self).__init__()

    def configure(self, masterappconfig):
        return (None, None)

    def templatelocation(self):
        """Provide name of template file with absolute path"""
        dir = path.dirname(path.abspath(inspect.getfile(inspect.currentframe())))
        return path.join(dir, 'wrapperNotebookTemplate.py.template')
                                                   
    def wrapper(self, regexp, version, timeout, kernel):
        """Write a wrapper Python script that executes the notebooks"""
        wrapperscript = FileUtils.loadScript(self.templatelocation(), '')

        wrapperscript = wrapperscript.replace('###NBFILES###', str(regexp))
        wrapperscript = wrapperscript.replace('###VERSION###', str(version))
        wrapperscript = wrapperscript.replace('###TIMEOUT###', str(timeout))
        wrapperscript = wrapperscript.replace('###KERNEL###', str(kernel))
        wrapperscript = wrapperscript.replace('###UUID###', str(uuid.uuid4()))

        logger.debug('Script to run on worker node\n' + wrapperscript)
        scriptName = "notebook_wrapper_generated.py"
        runScript = FileBuffer(scriptName, wrapperscript, executable=1)

        return runScript
    
    def unprepare(self, force=False):
        """
        Revert a Notebook application back to its unprepared state.
        """
        logger.debug('Running unprepare in Notebook app')
        if self.is_prepared is not None:
            self.decrementShareCounter(self.is_prepared)
            self.is_prepared = None
        self.hash = None

    def prepare(self, force=False):
        """
        This writes the wrapper script for the Notebook application.

        """
        if force:
            self.unprepare()
    
        if (self.is_prepared is not None):
            raise ApplicationPrepareError('%s application has already been prepared. Use prepare(force=True) to prepare again.' % getName(self))

        logger.info('Preparing %s application.' % getName(self))
        self.is_prepared = ShareDir()
        logger.info('Created shared directory: %s' % (self.is_prepared.name))
    
        # Prevent orphaned shared directories
        try:
            self.checkPreparedHasParent(self)

            script = self.wrapper(self.regexp,self.version, self.timeout, self.kernel)
            logger.debug("Creating: %s" % path.join(self.getSharedPath(), script.name))
            script.create(path.join(self.getSharedPath(), script.name))

            self.post_prepare()

        except Exception as err:
            self.unprepare()
            raise

        return 1


class NotebookRTHandler(IRuntimeHandler):
    """Empty runtime handler for notebooks"""

    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):

        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
        return StandardJobConfig('python', None, ['notebook_wrapper_generated.py'], None, None, [app.is_prepared.path()])


allHandlers.add('Notebook', 'LSF', NotebookRTHandler)
allHandlers.add('Notebook', 'Local', NotebookRTHandler)
allHandlers.add('Notebook', 'PBS', NotebookRTHandler)
allHandlers.add('Notebook', 'SGE', NotebookRTHandler)
allHandlers.add('Notebook', 'Condor', NotebookRTHandler)
allHandlers.add('Notebook', 'LCG', NotebookRTHandler)
allHandlers.add('Notebook', 'gLite', NotebookRTHandler)
allHandlers.add('Notebook', 'TestSubmitter', NotebookRTHandler)
allHandlers.add('Notebook', 'Interactive', NotebookRTHandler)
allHandlers.add('Notebook', 'Batch', NotebookRTHandler)
allHandlers.add('Notebook', 'Remote', NotebookRTHandler)
allHandlers.add('Notebook', 'CREAM', NotebookRTHandler)
allHandlers.add('Notebook', 'ARC', NotebookRTHandler)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
from copy import deepcopy
import tempfile
import fnmatch
from GangaCore.Core.exceptions import GangaException
from GangaCore.GPIDev.Lib.Dataset import GangaDataset
from GangaCore.GPIDev.Schema import GangaFileItem, SimpleItem, Schema, Version, ComponentItem
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.Utility.Config import getConfig, ConfigError
from GangaDirac.Lib.Files.DiracFile import DiracFile
import GangaCore.Utility.logging
import GangaLHCb.Lib.LHCbDataset
from .LHCbDatasetUtils import isLFN, isPFN, isDiracFile, strToDataFile, getDataFile
from GangaCore.GPIDev.Base.Proxy import isType, stripProxy, getName
from GangaCore.GPIDev.Lib.Job.Job import Job, JobTemplate
from GangaDirac.Lib.Backends.DiracUtils import get_result
from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList, makeGangaListByRef
from GangaCore.GPIDev.Adapters.IGangaFile import IGangaFile
logger = GangaCore.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

class LHCbCompressedFileSet(GangaObject):
    '''A class for handling sets of files. This stores the common start of the lfns,
    a list of all the non-common ends and a list of the metadata for each file.

    The metadata is stored as a list of tuples of numbers. If it is filled the order is
    (Luminosity, EvtStat, RunNo, TCK)
    '''

    schema = {}
    schema['lfn_prefix'] = SimpleItem(defvalue = None, typelist = ['str', None], doc = 'The common starting path of the LFN')
    schema['suffixes'] = SimpleItem(defvalue = [], typelist = [GangaList, 'str'], sequence=1, doc = 'The individual end of each LFN')
    _schema = Schema(Version(3, 0), schema)
    _exportmethods = ['__len__', 'getLFNs', 'getLFN']
    def __init__(self, files=None, lfn_prefix=None):
        super(LHCbCompressedFileSet, self).__init__()
        if lfn_prefix:
            self.lfn_prefix = lfn_prefix
            self.suffixes = [(_f) for _f in files]
        elif files:
            self.lfn_prefix = ''
            if not isType(files, [str, list, tuple, GangaList]):
                raise GangaException("Incorrect type %s passed to LHCbCompressedFileSet" % type(files))
            if isType(files, [list, tuple, GangaList]):
                commonpath = os.path.commonpath(files)
                if commonpath == '/':
                    commonpath = ''
                suffixes = [_lfn.replace(commonpath, '', 1) for _lfn in files]
                self.lfn_prefix = commonpath
                self.suffixes = suffixes
            else:
                self.files.append(files)

    def __len__(self):
        return len(self.suffixes)

    def getLFNs(self):
        '''Return a list of all the LFNs contained in this set'''
        lfns = [self.lfn_prefix + _suffix for _suffix in self.suffixes]
        return lfns

    def getLFN(self, i):
        '''Get a siingle LFN from the set'''
        new_lfn = self.lfn_prefix + self.suffixes[i]
        return new_lfn


class LHCbCompressedDataset(GangaDataset):

    '''Class for handling LHCb data sets (i.e. inputdata for LHCb jobs).
    This is a version of LHCbDataset that should use less disk space.
    It should only be used with DiracFile objects, and is best constructed
    from a BKQuery.

    All of the usual methods for datasets can be used here (extend, union, difference etc).
    These also work if the other dataset is a regular LHCbDataset.

    The LHCbCompressedDataset furthermore offers the ability to store some metadata about
    the files in it, i.e. Luminosity, EvtStat, Run no, and TCK

    For this dataset everything revolves around the LFN rather than individual file objects
    '''
    schema = {}
    docstr = 'List of DiracFile objects'
    schema['files'] = SimpleItem(defvalue=[], typelist=[LHCbCompressedFileSet], sequence=1, doc='A list of lists of the file suffixes')
    schema['XMLCatalogueSlice'] = GangaFileItem(defvalue=None, doc='Use contents of file rather than generating catalog.')
    schema['persistency'] = SimpleItem(defvalue=None, typelist=['str', 'type(None)'], doc='Specify the dataset persistency technology')
    schema['credential_requirements'] = ComponentItem('CredentialRequirement', defvalue=None)
    schema['depth'] = SimpleItem(defvalue = 0, doc='Depth')
    _schema = Schema(Version(3, 0), schema)
    _category = 'datasets'
    _name = "LHCbCompressedDataset"
    _exportmethods = ['getReplicas', '__len__', '__getitem__', '__iter__', '__next__', 'replicate',
                      'append', 'extend', 'getCatalog', 'optionsString', 'getFileNames', 'getFilenameList',
                      'getLFNs', 'getFullFileNames', 'getFullDataset', 'hasLFNs',
                      'difference', 'isSubset', 'isSuperset', 'intersection',
                      'symmetricDifference', 'union', 'bkMetadata', 'getMetadata',
                      'getLuminosity', 'getEvtStat', 'getRunNumbers', 'isEmpty', 'getPFNs'] 

    def __init__(self, files=None, metadata = None, persistency=None, depth=0, fromRef=False):
        super(LHCbCompressedDataset, self).__init__()
        self.files = []
        #if files is an LHCbDataset

        if files and isType(files, GangaLHCb.Lib.LHCbDataset.LHCbDataset):
            newset = LHCbCompressedFileSet(files.getLFNs())
            self.files.append(newset)
        #if files is an LHCbCompressedDataset
        if files and isType(files, LHCbCompressedDataset):
            self.files.extend(files.files)
        #if files is just a string
        if files and isType(files, str):
            newset = LHCbCompressedFileSet(files)
            self.files.append(newset)
        #if files is a single DiracFile
        if files and isType(files, DiracFile):
            newset = LHCbCompressedFileSet(files.lfn)
            self.files.append(newset)
        #if files is a single LHCbCompressedFileSet
        if files and isType(files, LHCbCompressedFileSet):
            self.files.append(files)
        #if files is a list
        if files and isType(files, [list, GangaList]):
            #Is it a list of strings? Then it may have been produced from the BKQuery so pass along the metadata as well
            if isType(files[0], str):
                newset = LHCbCompressedFileSet(files)
                self.files.append(newset)
            #Is it a list of DiracFiles?
            if isType(files[0], DiracFile):
                lfns = []
                for _df in files:
                    lfns.append(_df.lfn)
                newset = LHCbCompressedFileSet(lfns)
                self.files.append(newset)
            #Is it a list of file sets?
            if isType(files[0], LHCbCompressedFileSet):
                    self.files.extend(files)

        self.files._setParent(self)
        self.persistency = persistency
        self.current = 0
        self.total = self._totalNFiles()
        logger.debug("Dataset Created")


    def _location(self, i):
        '''Figure out where a file of index i is. Returns the subset no and the location within that subset'''
        setNo = 0
        fileTotal = 0
        while fileTotal < i+1 and setNo < len(self.files):
            fileTotal += len(self.files[setNo])
            setNo += 1
        if fileTotal < i:
            return -1, -1
        setNo = setNo - 1
        fileTotal = fileTotal - len(self.files[setNo])
        setLocation = i - fileTotal
        return setNo, setLocation

    def _totalNFiles(self):
        '''Return the total no. of files in the dataset'''
        total = 0
        for _set in self.files:
            total += len(_set)
        return total

    def __len__(self):
        '''Redefine the __len__ function'''
        return self.total

    def __getitem__(self, i):
        '''Proivdes scripting (e.g. ds[2] returns the 3rd file) '''
        if type(i) == type(slice(0)):
            #We construct a list of all LFNs first. Not the most efficient but it allows us to use the standard slice machinery
            newLFNs = self.getLFNs()[i]
            #We define these here for future speed
            indexLen = len(newLFNs)
            #Now pull out the prefixes/suffixes
            setNo = 0
            step = 1
            #If we are going backwards start at the end
            if i.step and i.step < 0:
                step = -1
                setNo = len(self.files)-1
            currentPrefix = None
            #Iterate over the LFNs and find out where it came from
            ds = LHCbCompressedDataset()
            tempList = []
            j = 0
            while j < len(newLFNs):
                if newLFNs[j] in self.files[setNo].getLFNs():
                    tempList.append(newLFNs[j])
                    j += 1
                else:
                    if len(tempList) > 0:
                        ds.addSet(LHCbCompressedFileSet(tempList))
                    setNo += step
                    tempList = []
            ds.addSet(LHCbCompressedFileSet(tempList))
        else:
            #Figure out where the file lies
            setNo, setLocation = self._location(i)
            if setNo < 0 or i >= self._totalNFiles():
                logger.error("Unable to retrieve file %s. It is larger than the dataset size" % i)
                return None
            ds = DiracFile(lfn = self.files[setNo].getLFN(setLocation), credential_requirements = self.credential_requirements)
        return ds

    def __iter__(self):
        '''Fix the iterator'''
        self.current = 0
        return self

    def __next__(self):
        '''Fix the iterator'''
        if self.current == self.total:
            raise StopIteration
        else:
            self.current += 1
            return self[self.current-1]

    def addSet(self, newSet):
        '''Add a new FileSet to the dataset'''
        self.files.append(newSet)
        self.total = self._totalNFiles()

    def getFileNames(self):
        'Returns a list of the names of all files stored in the dataset'
        names = []
        for _lfn in self.getLFNs():
            names.append(_lfn)
        return names

    def getFilenameList(self):
        'Return a list of filenames to be created as input.txt on the WN. These will be the PFNs'
        #We know these are DiracFiles so collate the LFNs and get the accessURLs together
        from GangaDirac.Lib.Backends.DiracUtils import getAccessURLs
        fileList = getAccessURLs(self.getLFNs())
        return fileList

    def getReplicas(self):
        'Returns the replicas for all files in the dataset.'
        lfns = self.getLFNs()
        cmd = 'getReplicas(%s)' % str(lfns)
        result = get_result(cmd, 'LFC query error. Could not get replicas.')
        return result['Successful']

    def extend(self, other, unique=False):
        '''Extend the dataset. If unique, then only add files which are not
        already in the dataset. You may extend with another LHCbCompressedDataset,
        LHCbDataset, DiracFile or a list of string of LFNs'''

        if isType(other, LHCbCompressedDataset):
            self.files.extend(other.files)
        elif isType(other, GangaLHCb.Lib.LHCbDataset.LHCbDataset):
            lfns = other.getLFNs()
            self.files.append(LHCbCompressedFileSet(lfns))
        elif isType(other, DiracFile):
            self.files.append(LHCbCompressedFileSet(other.lfn))
        elif isType(other, [list, tuple, GangaList]):
            self.files.append(LHCbCompressedFileSet(other))
        else:
            logger.error("Cannot add object of type %s to an LHCbCompressedDataset" % type(other))
        self.total = self._totalNFiles()

    def getLFNs(self):
        '''Returns a list of all LFNs (by name) stored in the dataset.'''
        lfns = []
        if not self:
            return lfns
        for fileset in self.files:
            lfns.extend(fileset.getLFNs())
        logger.debug("Returning #%s LFNS" % str(len(lfns)))
        return lfns

    def getMetadata(self):
        '''Returns a list of all the metadata'''
        from GangaLHCb.Lib.Backends.Dirac import getLFNMetadata
        return getLFNMetadata(self.getLFNs())

    def getLuminosity(self):
        '''Returns the total luminosity'''
        lumi = 0
        mets = self.getMetadata()
        for _f in mets.keys():
            lumi += mets[_f]['Luminosity']
        return lumi

    def getEvtStat(self):
        '''Returns the total events'''
        evtStat = 0
        mets = self.getMetadata()
        for _f in mets.keys():
            evtStat += mets[_f]['EventStat']
        return evtStat

    def getRunNumbers(self):
        '''Returns a list of the run numbers'''
        runNos = []
        mets = self.getMetadata()
        for _f in mets.keys():
            runNos.append(mets[_f]['RunNumber'])
        return runNos

    def isEmpty(self):
        '''Does this contain files'''
        return not len(self)>0

    def hasLFNs(self):
        '''Does it contain LFNs'''
        return not self.isEmpty()

    def getPFNs(self):
        'Returns a list of all PFNs (by name) stored in the dataset.'
        return self.getFilenameList()

    def getFullFileNames(self):
        'Returns all file names with LFN prepended.'
        names = []
        for _lfn in self.getLFNs():
                names.append('LFN:%s' % _lfn)
        return names

    def getFullDataset(self):
        '''Returns an LHCb dataset'''
        ds = GangaLHCb.Lib.LHCbDataset.LHCbDataset(persistency = self.persistency)
        lfns = self.getLFNs()
        for _lfn in lfns:
            ds.extend(DiracFile(lfn = _lfn))
        return ds

    def getCatalog(self, site=''):
        '''Generates an XML catalog from the dataset (returns the XML string).
        Note: site defaults to config.LHCb.LocalSite
        Note: If the XMLCatalogueSlice attribute is set, then it returns
              what is written there.'''
        if hasattr(self.XMLCatalogueSlice, 'name'):
            if self.XMLCatalogueSlice.name:
                f = open(self.XMLCatalogueSlice.name)
                xml_catalog = f.read()
                f.close()
                return xml_catalog
        if not site:
            site = getConfig('LHCb')['LocalSite']
        lfns = self.getLFNs()
        depth = self.depth
        tmp_xml = tempfile.NamedTemporaryFile(suffix='.xml')
        cmd = 'getLHCbInputDataCatalog(%s,%d,"%s","%s")' \
              % (str(lfns), depth, site, tmp_xml.name)
        result = get_result(cmd, 'LFN->PFN error. XML catalog error.')
        xml_catalog = tmp_xml.read()
        tmp_xml.close()
        return xml_catalog

    def optionsString(self, file=None):
        'Returns the Gaudi-style options string for the dataset (if a filename' \
            ' is given, the file is created and output is written there).'
        if not self or len(self) == 0:
            return ''
        snew = ''
        if self.persistency == 'ROOT':
            snew = '\n#new method\nfrom GaudiConf import IOExtension\nIOExtension(\"%s\").inputFiles([' % self.persistency
        elif self.persistency == 'POOL':
            snew = '\ntry:\n    #new method\n    from GaudiConf import IOExtension\n    IOExtension(\"%s\").inputFiles([' % self.persistency
        elif self.persistency is None:
            snew = '\ntry:\n    #new method\n    from GaudiConf import IOExtension\n    IOExtension().inputFiles(['
        else:
            logger.warning(
                "Unknown LHCbCompressedDataset persistency technology... reverting to None")
            snew = '\ntry:\n    #new method\n    from GaudiConf import IOExtension\n    IOExtension().inputFiles(['

        sold = '\nexcept ImportError:\n    #Use previous method\n    from Gaudi.Configuration import EventSelector\n    EventSelector().Input=['
        sdatasetsnew = ''
        sdatasetsold = ''

        dtype_str_default = getConfig('LHCb')['datatype_string_default']
        dtype_str_patterns = getConfig('LHCb')['datatype_string_patterns']
        for _f in self.getLFNs():
            dtype_str = dtype_str_default
            for this_str in dtype_str_patterns:
                matched = False
                for pat in dtype_str_patterns[this_str]:
                    if fnmatch.fnmatch(_f, pat):
                        dtype_str = this_str
                        matched = True
                        break
                if matched:
                    break
            sdatasetsnew += '\n        '
            sdatasetsold += '\n        '
            #f is always a DiracFile
            sdatasetsnew += """ \"LFN:%s\",""" % _f
            sdatasetsold += """ \"DATAFILE='LFN:%s' %s\",""" % (_f, dtype_str)

        if sdatasetsold.endswith(","):
            if self.persistency == 'ROOT':
                sdatasetsnew = sdatasetsnew[:-1] + """\n], clear=True)"""
            else:
                sdatasetsnew = sdatasetsnew[:-1] + """\n    ], clear=True)"""
            sdatasetsold = sdatasetsold[:-1]
            sdatasetsold += """\n    ]"""
        if(file):
            f = open(file, 'w')
            if self.persistency == 'ROOT':
                f.write(snew)
                f.write(sdatasetsnew)
            else:
                f.write(snew)
                f.write(sdatasetsnew)
                f.write(sold)
                f.write(sdatasetsold)
            f.close()
        else:
            if self.persistency == 'ROOT':
                return snew + sdatasetsnew
            else:
                return snew + sdatasetsnew + sold + sdatasetsold

    def _checkOtherFiles(self, other ):
        if isType(other, GangaList) or isType(other, []):
            other_files = other
        elif isType(other, LHCbCompressedDataset):
            other_files = other.getLFNs()
        elif isType(other, GangaLHCb.Lib.LHCbDataset.LHCbDataset):
            other_files = other.getLFNs()
        else:
            raise GangaException("Unknown type for difference")
        return other_files

    def difference(self, other):
        '''Returns a new data set w/ files in this that are not in other.'''
        other_files = self._checkOtherFiles(other)
        files = set(self.getLFNs()).difference(other_files)
        data = LHCbCompressedDataset(list(files))
        return data

    def isSubset(self, other):
        '''Is every file in this data set in other?'''
        other_files = self._checkOtherFiles(other)
        return set(self.getLFNs()).issubset(other_files)

    def isSuperset(self, other):
        '''Is every file in other in this data set?'''
        other_files = self._checkOtherFiles(other)
        return set(self.getLFNs()).issuperset(other_files)

    def symmetricDifference(self, other):
        '''Returns a new data set w/ files in either this or other but not
        both.'''
        other_files = self._checkOtherFiles(other)
        files = set(self.getLFNs()).symmetric_difference(other_files)
        data = LHCbCompressedDataset(list(files))
        return data

    def intersection(self, other):
        '''Returns a new data set w/ files common to this and other.'''
        other_files = self._checkOtherFiles(other)
        files = set(self.getLFNs()).intersection(other_files)
        data = LHCbCompressedDataset(list(files))
        return data

    def union(self, other):
        '''Returns a new data set w/ files from this and other.'''
        other_files = self._checkOtherFiles(other)
        files = set(self.getLFNs()).union(other_files)
        data = LHCbCompressedDataset(list(files))
        return data

    def bkMetadata(self):
        'Returns the bookkeeping metadata for all LFNs. '
        logger.info("Using BKQuery(bkpath).getDatasetMetadata() with bkpath=the bookkeeping path, will yeild more metadata such as 'TCK' info...")
        cmd = 'bkMetaData(%s)' % self.getLFNs()
        b = get_result(cmd, 'Error removing getting metadata.')
        return b





# generate job status transition graph
# usage:
#    1) run ganga and this file as a script
# or 2) cd ganga/python, run python interactively and execfile(this_file)

ARC_LABELS = True
STYLED_EDGES = True
DEBUG = False

from Ganga.GPIDev.Lib.Job import Job
import os

g = Job.status_graph
initial_states = Job.initial_states
transient_states = Job.transient_states

import os.path
dot_file = os.path.abspath('ganga_job_stat.dot')
out_type = 'gif'
out_file = dot_file.replace('.dot','.'+out_type)

def debug(s):
    if DEBUG:
        print 'DEBUG:',s

f = file(dot_file,'w')
print >> f, 'digraph JobStatus {'
for node in g:
    debug('src state: %s'%node)
    for dest in g[node]:
        debug('dest state: %s'%dest)        
        LAB = []
        label = g[node][dest].transition_comment
        if ARC_LABELS:
            LAB.append('label="%s"'%label)
            LAB.append('fontsize=8')
        if STYLED_EDGES:
            if label.find('force') != -1:
                LAB.append('style=dotted bold')
        if LAB:
            LAB = '['+','.join(LAB)+']'
        print >>f, '%s -> %s %s;' % (node,dest,LAB)

print >>f,"__start [shape=point]"

for node in initial_states:
    print >>f, '__start -> %s;'%node

for node in transient_states:
    print >>f, '%s [style=filled]'%node
    
print >>f, '}'
f.close()

print 'created', dot_file

#graphviz_top = '/afs/cern.ch/sw/lcg/external/graphviz/1.9/rh73_gcc32/'
#os.system('export LD_LIBRARY_PATH=%s.lib/graphviz:$LD_LIBRARY_PATH; %s/bin/dot -T%s %s -o%s'% (graphviz_top,graphviz_top,out_type, dot_file, out_file))

os.system('dot -T%s %s -o%s'% (out_type, dot_file, out_file))
print 'created', out_file



from GangaCore.GPIDev.Schema import *
from GangaCore.GPIDev.Lib.Tasks.common import *
from GangaCore.GPIDev.Lib.Tasks.ITransform import ITransform
from GangaCore.GPIDev.Lib.Job.Job import JobError
from GangaCore.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.GPIDev.Lib.Tasks.ITransform import ITransform
from GangaCore.GPIDev.Lib.Tasks.TaskLocalCopy import TaskLocalCopy
from GangaCore.Utility.logging import getLogger
from .ND280Unit_CSVEvtList import ND280Unit_CSVEvtList
from GangaND280.ND280Dataset.ND280Dataset import ND280LocalDataset
from GangaND280.ND280Splitter import splitCSVFile
import GangaCore.GPI as GPI

import os

logger = getLogger()

class ND280Transform_CSVEvtList(ITransform):
   _schema = Schema(Version(1,0), dict(list(ITransform._schema.datadict.items()) + list({
        'nbevents' : SimpleItem(defvalue=-1,doc='The number of events for each unit'),
    }.items())))

   _category = 'transforms'
   _name = 'ND280Transform_CSVEvtList'
   _exportmethods = ITransform._exportmethods + [ ]

   def __init__(self):
      super(ND280Transform_CSVEvtList,self).__init__()

   def createUnits(self):
      """Create new units if required given the inputdata"""
      
      # call parent for chaining
      super(ND280Transform_CSVEvtList,self).createUnits()
      
      # Look at the application schema and check if there is a csvfile variable
      try:
        csvfile = self.application.csvfile
      except AttributeError:
        logger.error('This application doesn\'t contain a csvfile variable. Use another Transform !')
        return

      subsets = splitCSVFile(self.application.csvfile, self.nbevents)

      for s,sub in enumerate(subsets):
        
        # check if this data is being run over by checking all the names listed
        ok = False
        for unit in self.units:
          if unit.subpartid == s:
            ok = True

        if ok:
          continue
        # new unit required for this dataset
        unit = ND280Unit_CSVEvtList()
        unit.name = "Unit %d" % len(self.units)
        unit.subpartid = s
        unit.eventswanted = sub
        unit.inputdata = self.inputdata[0]
        self.addUnitToTRF( unit )


   def createChainUnit( self, parent_units, use_copy_output = True ):
      """Create a chained unit using the output data from the given units"""

      # check all parent units for copy_output
      copy_output_ok = True
      for parent in parent_units:
         if not parent.copy_output:
            copy_output_ok = False

      # all parent units must be completed so the outputfiles are filled correctly
      for parent in parent_units:
         if parent.status != "completed":
               return None

      if not use_copy_output or not copy_output_ok:
         unit = ND280Unit_CSVEvtList()
         unit.inputdata = ND280LocalDataset()
         for parent in parent_units:
            # loop over the output files and add them to the ND280LocalDataset - THIS MIGHT NEED SOME WORK!
            job = GPI.jobs(parent.active_job_ids[0])
            for f in job.outputfiles:
               # should check for different file types and add them as appropriate to the dataset
               # self.inputdata (== TaskChainInput).include/exclude_file_mask could help with this
               # This will be A LOT easier with Ganga 6.1 as you can easily map outputfiles -> inputfiles!
               unit.inputdata.names.append( os.path.join( job.outputdir, f.namePattern ) )
      else:

         unit = ND280Unit_CSVEvtList()
         unit.inputdata = ND280LocalDataset()

         for parent in parent_units:
            # unit needs to have completed and downloaded before we can get file list
            if parent.status != "completed":
               return None

            # we should be OK so copy all output to the dataset
            for f in parent.copy_output.files:
               unit.inputdata.names.append( os.path.join( parent.copy_output.local_location, f ) )
         
      return unit
   



#! /bin/env python

"""
WARNING: urllib2 does not validate SSL certificates before python 3.2.2.

This is insecure (the USERNAME and PASSWORD could theoretically be snooped by a 
man-in-the-middle attack).

@TODO add cafile parameter when python upgraded to 3.2.2

-----

bash-3.2$ ./post_status.py -h
usage: ./post_status.py [options] MONDIR RUN SUBRUN TRIGTYPE STAGE
       ./post_status.py --site=scinet --result=1 --time=1577 --read=5062 --written=310 production004/B/rdp/ND280/00005000_00005999 5012 6 spill reco
       ./post_status.py -h
       
       Record metadata about a processing job in the processingstatus db.
       
       Intended to be called by check_mon_info.pl.

       If a server problem occurs, this script retries up to MAX_RETRIES times.

       EXIT CODES

       0 - Request succeeded
       1 - Unknown problem
       2 - Failed to parse command line (detected by this script)
       3 - Command line argument(s) invalid or incorrect type (detected by the database server)
       4 - Server problem (http response code != 201)

arguments:
       MONDIR - container for a set of metadata records
       should be initialized following the Production Directory Structure
       http://www.t2k.org/nd280/datacomp/howtoaccessdata/directorystructure

       RUN - integer                                               --| Composite Primary 
       SUBRUN - integer                                              | Key that
       TRIGTYPE - string                                             | identifies a
       https://neut00.triumf.ca/t2k/processing/status/trigtypes      | processing job
       STAGE - string                                                | within a mondir 
       https://neut00.triumf.ca/t2k/processing/status/stages       --|

options:
  -h, --help         show this help message and exit
  --site=SITE        Site (string)
  --result=RESULT    Result (integer) previously registered at https://neut00.
                     triumf.ca/t2k/processing/status/result_codes
  --time=TIME        Wallclock time in seconds (integer)
  --read=READ        Events read from input file (integer)
  --written=WRITTEN  Events written to output file (integer)

-----

It would be better to communicate through message queues, rather than 
posting via HTTP directly to the application (this would make the posting of 
statuses more reliable, scalable, flexible and require less custom code).

However, this would require additional infrastructure:

 * an AMQP message broker (such as RabbitMQ or better yet, ZeroMQ)
 * an AMQP library (such as Carrot)
 * a data serialization format (such as json or msgpack)

These dependencies would make the system harder to maintain.
"""

#APP_ROOT='http://neut07.triumf.ca:4000/'
#APP_ROOT='https://neut00.triumf.ca/t2k/processing/status/'
APP_ROOT='https://nd280web.nd280.org/t2k/processing/status/'

USERNAME='procstatupdater'
PASSWORD='2aT0n8Ra'

MAX_RETRIES = 20

#####

import optparse

import urllib.parse, urllib.request, urllib.parse, urllib.error, urllib.request, urllib.error, urllib.parse

import random, time

import sys

#import pdb

try:
    # for python 2.6+
    import json
except ImportError:
    # for python < 2.6
    class json:
        @staticmethod
        def dumps(obj):
            if obj.__class__ != dict:
                raise Exception("Only dict serialization is implemented")
            o = []
            for (k,v) in list(obj.items()):
                if v is None:
                    o.append( '"%s": null' % (k, ) )
                elif v.__class__ == str:
                    o.append( '"%s": "%s"' % (k, v) )
                elif v.__class__ in [int, float]:
                    o.append( '"%s": %s' % (k, v) )
                else:
                    o.append( '"%s": %s' % (k, json.dumps(v)) )
            return "{%s}" % (', '.join(o), )

#####

def record(mondir, job, attributes):

    #blech - we have to subclass to handle responses with 201 status code
    class CustomHTTPErrorProcessor(urllib2.HTTPErrorProcessor):
        handler_order = 1000  # after all other processing
        def http_response(self, request, response):
            code, msg, hdrs = response.code, response.msg, response.info()
            if code != 201:
                response = self.parent.error('http', request, response, code, msg, hdrs)
            return response
        https_response = http_response

    query_string = urllib.parse.urlencode(job)
    mondir_url = urllib.parse.urljoin(APP_ROOT, mondir) + '?' + query_string
    attributes_json = json.dumps(attributes)

    print("POST %s\n%s" % (mondir_url, attributes_json))

    # build auth handler
    pswd_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
    pswd_mgr.add_password(None, APP_ROOT, USERNAME, PASSWORD)
    auth_handler = urllib.request.HTTPBasicAuthHandler(pswd_mgr)

    # build opener
    opener = urllib.request.build_opener(auth_handler, CustomHTTPErrorProcessor())

    attempt = 0	
    while True:
        try:
            response = opener.open( urllib.request.Request(mondir_url, attributes_json, {'Accept':'text/plain'}) )
            # (blech)
	    #pdb.set_trace()
            #(Pdb) dir(response)
            #['__doc__', '__init__', '__iter__', '__module__', '__repr__', 'close', 'code', 'fileno', 'fp', 'geturl', 'headers', 'info', 'msg', 'next', 'read', 'readline', 'readlines', 'url']
            print("%s %s" % (response.code, response.msg))
            print(response.read())
            break
        except urllib.error.HTTPError as e:
            if e.code == 400:
                print(e)
                print(e.read())
                print("Command line argument(s) invalid (detected by the database server)")
                return 3
            elif e.code == 404:
                print(e)
                print(e.read())
                print("MONDIR does not exist (detected by database server)")
                return 3
            print("Error: %s %s\n%s" % ( e.code, e.msg, e.read() ))
            if attempt < MAX_RETRIES:
                sleep_time = random.random() * 10
                print("--> Retrying in %s seconds" % (sleep_time, ))
                time.sleep(sleep_time)
                attempt += 1
                continue
            else:
                raise

#####

def parse_cmdline():
    usage = """./%prog [options] MONDIR RUN SUBRUN TRIGTYPE STAGE
       ./%prog --site=scinet --result=1 --time=1577 --read=5062 --written=310 production004/B/rdp/ND280/00005000_00005999 5012 6 spill reco
       ./%prog -h
       
       Record metadata about a processing job in the processingstatus db.
       
       Intended to be called by check_mon_info.pl.

       If a server problem occurs, this script retries up to MAX_RETRIES times.

       EXIT CODES

       0 - Request succeeded
       1 - Unknown problem
       2 - Failed to parse command line (detected by this script)
       3 - Command line argument(s) invalid or incorrect type (detected by the database server)
       4 - Server problem (http response code != 201)

arguments:
       MONDIR - container for a set of metadata records
       should be initialized following the Production Directory Structure
       http://www.t2k.org/nd280/datacomp/howtoaccessdata/directorystructure

       RUN - integer                                               --| Composite Primary 
       SUBRUN - integer                                              | Key that
       TRIGTYPE - string                                             | identifies a
       https://neut00.triumf.ca/t2k/processing/status/trigtypes      | processing job
       STAGE - string                                                | within a mondir 
       https://neut00.triumf.ca/t2k/processing/status/stages       --|"""

    p = optparse.OptionParser(usage=usage)

    p.add_option("--site", type='string', help='Site (string)')
    p.add_option("--result", type='int', help='Result (integer) previously registered at https://neut00.triumf.ca/t2k/processing/status/result_codes')
    p.add_option("--time", type='int', help='Wallclock time in seconds (integer)')
    p.add_option("--read", type='int', help='Events read from input file (integer)')
    p.add_option("--written", type='int', help='Events written to output file (integer)')

    (opts, args) = p.parse_args()

    if len(args) != 5:
        p.error("incorrect number of arguments")

    job = { 'run':int(args[1]),
            'subrun':int(args[2]),
            'trigtype':args[3],
            
            'stage':args[4] }
            
    attributes = { 'site':opts.site,
                 'result':opts.result,
                 'time':opts.time,
                 'read':opts.read,
                 'written':opts.written }

    return (args[0], job, attributes)

#####

if __name__ == '__main__':
    (mondir, job, attributes) = parse_cmdline()
    
    try:
        record(mondir, job, attributes)
    except urllib.error.HTTPError as e:
        print("Error: The HTTP request failed. (%s)" % (str(e), ))
        sys.exit(4)

    sys.exit(0)



from GangaCore.Core.exceptions import SplitterError
from GangaGaudi.Lib.Splitters.GaudiInputDataSplitter import GaudiInputDataSplitter
from GangaCore.Core.exceptions import SplitterError
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.files import expandfilename
from GangaCore.GPIDev.Base.Proxy import stripProxy, isType, getName
import GangaCore.Utility.logging
from GangaCore.GPIDev.Lib.Job import Job
from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.Credentials.DiracProxy import DiracProxy
import os
import copy
import pickle

from GangaCore.GPIDev.Adapters.IGangaFile import IGangaFile
from GangaCore.GPIDev.Base.Filters import allComponentFilters
from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
from GangaCore.GPIDev.Credentials import require_credential
from GangaLHCb.Lib.LHCbDataset.LHCbDataset import LHCbDataset
from GangaLHCb.Lib.LHCbDataset.LHCbCompressedDataset import LHCbCompressedDataset
logger = GangaCore.Utility.logging.getLogger()


def getBackend():
    """
    Returns the backend which is to be used by the splitter based upon the LHCb config
    """
    lhcbConfig = getConfig('LHCb')
    return lhcbConfig['SplitByFilesBackend']


class SplitByFiles(GaudiInputDataSplitter):

    """Splits a job into sub-jobs by partitioning the input data

    SplitByFiles can be used to split a job into multiple subjobs, where
    each subjob gets an unique subset of the inputdata files.
    """
    _name = 'SplitByFiles'
    _schema = Schema(Version(1, 0), {
        'filesPerJob': SimpleItem(defvalue=10,
                                  doc='Number of files per subjob',
                                  typelist=['int']),

        'maxFiles': SimpleItem(defvalue=None,
                               doc='Maximum number of files to use in a masterjob (None = all files)',
                               typelist=['int', 'type(None)']),

        'bulksubmit': SimpleItem(defvalue=False,
                                 doc='determines if subjobs are split '
                                 'server side in a "bulk" submission or '
                                 'split locally and submitted individually'),

        'ignoremissing': SimpleItem(defvalue=False,
                                    doc='Skip LFNs if they are not found '
                                    'in the LFC. This option is only used if'
                                    'jobs backend is Dirac'),

        'splitterBackend': SimpleItem(defvalue=getBackend(),
                                      doc='name of the backend algorithm to use for splitting',
                                      typelist=['str'], protected=1, visitable=0),
        'credential_requirements': ComponentItem('CredentialRequirement', defvalue=DiracProxy),
        })

    _exportmethods = ['split']

    def _attribute_filter__set__(self, name, value):
        if name is 'filesPerJob':
            if value > 100:
                logger.warning('filesPerJob exceeded DIRAC maximum')
                logger.warning('DIRAC has a maximum dataset limit of 100.')
                logger.warning(
                    'BE AWARE!... will set it to this maximum value at submit time if backend is Dirac')
        return value

    def _create_subjob(self, job, dataset):
        logger.debug("_create_subjob")
        datatmp = []

        logger.debug("dataset size: %s" % str(len(dataset)))
        #logger.debug( "dataset: %s" % str(dataset) )

        if isinstance(dataset, LHCbDataset):
            for i in dataset:
                if isType(i, DiracFile):
                    datatmp.append(i)
                else:
                    logger.error("Unkown file-type %s, cannot perform split with file %s" % (type(i), str(i)))
                    from GangaCore.Core.exceptions import GangaException
                    raise GangaException("Unkown file-type %s, cannot perform split with file %s" % (type(i), str(i)))
        elif isinstance(dataset, (list, GangaList)):
            for this_file in dataset:
                if type(this_file) is str:
                    datatmp.append(allComponentFilters['gangafiles'](this_file, None))
                elif isType(this_file, IGangaFile):
                    datatmp.append(this_file)
                else:
                    logger.error("Unexpected type: %s" % str(type(this_file)))
                    logger.error("Wanted object to inherit from type: %s: %s" % (str(type(IGangaFile()))))
                    from GangaCore.Core.exceptions import GangaException
                    x = GangaException("Unknown(unexpected) file object: %s" % this_file)
                    raise x
        elif type(dataset) is str:
            datatmp.append(DiracFile(lfn=dataset))
        else:
            logger.error("Unkown dataset type, cannot perform split here")
            from GangaCore.Core.exceptions import GangaException
            logger.error("Dataset found: " + str(dataset))
            raise GangaException("Unkown dataset type, cannot perform split here")

        logger.debug("Creating new Job in Splitter")
        j = Job()
        logger.debug("Copying From Job")
        j.splitterCopy(stripProxy(job), ['splitter', 'subjobs', 'inputdata', 'inputsandbox', 'inputfiles', 'fqid', 'outputdir', 'master', 'merger', 'metadata', 'been_queued', 'parallel_submit', 'inputdir', 'non_copyable_outputfiles', 'id','status'])
        logger.debug("Unsetting Splitter")
        j.splitter = None
        logger.debug("Setting InputData")
        if isType(job.inputdata, LHCbCompressedDataset):
            newDs = LHCbCompressedDataset(datatmp[:])
        else:
            newDs = LHCbDataset(files=datatmp[:],
                                  persistency=self.persistency,
                                  depth=self.depth)
        j.inputdata = newDs
        logger.debug("Returning new subjob")
        return j

    @require_credential
    def _splitter(self, job, inputdata):
        """
        This is the main method used in splitting by inputdata for Dirac backends

        Args:
            job (Job): This is the master-job object which contains everything we need to split by inputdata
            inputdata (list): List of (DiracFile) objects which we're to use in the splitting

        Returns:
            outdata (generator): a splitter Generator

        """

        logger.debug("_splitter")

        if isType(inputdata, LHCbCompressedDataset):
            indata = inputdata.getFullDataset()
        else:
            indata = inputdata

        if indata is not None:

            self.depth = indata.depth
            self.persistency = indata.persistency
            self.XMLCatalogueSlice = indata.XMLCatalogueSlice

        else:

            self.depth = 0
            self.persistency = None
            self.XMLCatalogueSlice = None

        if stripProxy(job.backend).__module__.find('Dirac') > 0:
            bannedSites = []
            if 'BannedSites' in job.backend.settings:
                bannedSites = job.backend.settings['BannedSites']
            logger.debug("found Dirac backend")

            if self.filesPerJob > 100:
                self.filesPerJob = 100  # see above warning
            logger.debug("indata: %s " % str(indata))

            if self.splitterBackend == "GangaDiracSplitter":
                from GangaDirac.Lib.Splitters.GangaSplitterUtils import GangaDiracSplitter
                outdata = GangaDiracSplitter(indata,
                                             self.filesPerJob,
                                             self.maxFiles,
                                             self.ignoremissing)
            elif self.splitterBackend == "OfflineGangaDiracSplitter":
                from GangaDirac.Lib.Splitters.OfflineGangaDiracSplitter import OfflineGangaDiracSplitter
                outdata = OfflineGangaDiracSplitter(indata,
                                                    self.filesPerJob,
                                                    self.maxFiles,
                                                    self.ignoremissing,
                                                    bannedSites)
            elif self.splitterBackend == "splitInputDataBySize":
                from GangaLHCb.Lib.Splitters.LHCbSplitterUtils import DiracSizeSplitter
                outdata = DiracSizeSplitter(indata,
                                            self.filesPerJob,
                                            self.maxFiles,
                                            self.ignoremissing)
            elif self.splitterBackend == "splitInputData":
                indata = stripProxy(copy.deepcopy(inputdata))
                from GangaDirac.Lib.Splitters.SplitterUtils import DiracSplitter
                outdata = DiracSplitter(indata,
                                        self.filesPerJob,
                                        self.maxFiles,
                                        self.ignoremissing)
            else:
                raise SplitterError("Backend algorithm not selected!")

            logger.debug("outdata: %s " % str(outdata))
            return outdata
        #If we are not running the jobs on Dirac but are using DiracFiles we want some of the same checks
        elif all(isinstance(this_file, DiracFile) for this_file in indata):
            from GangaDirac.Lib.Splitters.OfflineGangaDiracSplitter import OfflineGangaDiracSplitter
            outdata = OfflineGangaDiracSplitter(indata,
                                              self.filesPerJob,
                                              self.maxFiles,
                                              self.ignoremissing)
            return outdata
        else:
            logger.debug("Calling Parent Splitter as not on Dirac")
            return super(SplitByFiles, self)._splitter(job, indata)

    @require_credential
    def split(self, job):
        """
        This is the main method which will split the given job

        Args:
            job(Job): This is the master job which will be used to split based upon the inputdata

        Return:
            split_return(list): This is the list of subjobs created from the input job
        """
        logger.debug("split")
        if self.maxFiles == -1:
            self.maxFiles = None
        if self.bulksubmit == True:
            if stripProxy(job.backend).__module__.find('Dirac') > 0:
                logger.debug("Returning []")
                return []
        split_return = super(SplitByFiles, self).split(job)
        logger.debug("split_return: %s" % split_return)
        return split_return




"""Simple script to replace @DATE@ and friends with real information.

Usage:  rewrite.py boilerplate.tex [VAR=value] ... <template >output
"""

import sys
import time


def get_info(fp):
    s = fp.read()

    d = {}
    start = s.find(r"\date{")
    if start >= 0:
        end = s.find("}", start)
        date = s[start+6:end]
        if date == r"\today":
            date = time.strftime("%B %d, %Y", time.localtime(time.time()))
        d["DATE"] = date
    return d
    

def main():
    s = sys.stdin.read()
    if "@" in s:
        # yes, we actully need to load the replacement values
        d = get_info(open(sys.argv[1]))
        for arg in sys.argv[2:]:
            name, value = arg.split("=", 1)
            d[name] = value
        start = 0
        while 1:
            start = s.find("@", start)
            if start < 0:
                break
            end = s.find("@", start+1)
            name = s[start+1:end]
            if name:
                value = d.get(name)
                if value is None:
                    start = end + 1
                else:
                    s = s[:start] + value + s[end+1:]
                    start = start + len(value)
            else:
                # "@@" --> "@"
                s = s[:start] + s[end:]
                start = end
    sys.stdout.write(s)


if __name__ == "__main__":
    main()



try:
    from urllib.request import urlopen
except:
    from urllib2 import urlopen

import subprocess
import os
import urllib
import os

def checkSingularity():
    """Check whether Singularity is installed and the current user has right to access

        Return value: True or False"""

    nullOutput = open(os.devnull, 'wb')
    returnCode = 1
    try:
        returnCode = subprocess.call(["singularity", "--version"], stdout=nullOutput, stderr=nullOutput)
    except:
        pass 
    if returnCode == 0 : return True
    return False


def checkDocker():
    """Check whether Docker is installed and the current user has right to access

        Return value: True or False"""

    nullOutput = open(os.devnull, 'wb')
    returnCode = 1
    try:
        returnCode = subprocess.call(["docker", "ps"], stdout=nullOutput, stderr=nullOutput)
    except:
        pass 
    if returnCode == 0 : return True
    return False


def checkUDocker(location='~'):
    """Check whether UDocker is installed and the current user has right to access

        Return value: True or False"""

    fname = os.path.join(os.path.expanduser(location),"udocker")
    nullOutput = open(os.devnull, 'wb')
    if (os.path.isfile(fname)):
        returnCode = subprocess.call([fname, "ps"], stdout=nullOutput, stderr=nullOutput)
        if (returnCode == 0):
            return True
    return False


def installUdocker(location='~'):
    """Download and install UDocker

        Return value: True (If Success) or False"""

    fname = os.path.join(os.path.expanduser(location),"udocker")
    url = "https://raw.githubusercontent.com/indigo-dc/udocker/master/udocker.py"

    import ssl
    context = ssl._create_unverified_context()
    try:
        with urlopen(url, context=context) as response, open(fname, 'wb') as out_file:
            data = response.read()
            out_file.write(data)
    except:
        returnCode = subprocess.check_call(['curl', '-k', url], stdout=open(fname, 'wb'))
        if (returnCode != 0):
            raise OSError('Error downloading uDocker')

    subprocess.call(["chmod", "u+rx", fname])
    udockerdir = os.path.join(location,'.udocker')
    os.environ['UDOCKER_DIR']=udockerdir
    os.makedirs(udockerdir)
    with open(os.path.join(udockerdir,'udocker.conf'), 'w') as fconfig:
        fconfig.write('http_insecure = True')
    returnCode = subprocess.call([fname, "install"])
    if (returnCode != 0):
        raise OSError('Error installing uDocker')
    print('UDocker Successfully installed')




from GangaCore.GPIDev.Base.Proxy import *
from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Schema import *


class LHCbTaskDummySplitter(ISplitter):
    _schema = Schema(Version(1, 0), {
        'orig_splitter': ComponentItem('splitters', defvalue=None, load_default=0, optional=1, doc='original splitter'),
    })
    _category = 'splitters'
    _name = 'LHCbTaskDummySplitter'
    _exportmethods = []
    _hidden = 1

    def __init__(self, splitter):
        super(LHCbTaskDummySplitter, self).__init__()
        self.orig_splitter = splitter

    def clone(self):
        return LHCbTaskDummySplitter(self.orig_splitter)

    def split(self, job):
        subjobs = self.orig_splitter.validatedSplit(job)
        for sj in subjobs:
            #sj.application.id = -1
            sj.name += ':%i' % (subjobs.index(sj))
        return subjobs



import subprocess
from datetime import datetime
import time
import uuid
from os import makedirs, path
from GangaCore.Core.exceptions import GangaException
from GangaCore.Runtime.GPIexport import exportToGPI
from GangaCore.Utility.files import expandfilename
from GangaCore.Utility.logging import getLogger
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.Core.exceptions import ApplicationPrepareError
from .PythonOptsCmakeParser import PythonOptsCmakeParser

logger = getLogger()

def gaudiPythonWrapper(sys_args, options, data_file, script_file):
    """
    Returns the script which is used by GaudiPython to wrap the job on the WN
    Args:
        sys_args
    """
    wrapper_script = """
from Gaudi.Configuration import *

import sys
sys.argv += %s

# Import extra Opts
importOptions('%s')

# Import data.py
importOptions('%s')

for script_file in %s:
    importOptions( script_file )

# Run the command
execfile('%s')
""" % (sys_args, options, data_file, repr(script_file[1:]), script_file[0])
    return wrapper_script

def getTimestampContent():
    """
    Returns a string containing the current time in a given format and a unique random uuid
    """
    fmt = '%Y-%m-%d-%H-%M-%S'
    return datetime.now().strftime(fmt) + '\n' + str(uuid.uuid4())

def addTimestampFile(given_path, fileName='__timestamp__'):
    """
    This creates a file in this directory given called __timestamp__ which contains the time so that the final file is unique
    I also add a unique UUID which reduces the risk of collisions between users
    Args:
        given_path (str): Path which we want to create the timestamp within
    """
    time_filename = path.join(given_path, fileName)
    logger.debug("Constructing: %s" % time_filename)
    with open(time_filename, 'a+') as time_file:
        time_file.write(getTimestampContent())

def getGaudiExecInputData(optsfiles, app):
    '''Returns a LHCbDataSet object from a list of options files. The
       optional argument extraopts will decide if the extraopts string inside
       the application is considered or not.

    Usage example:
    # Get the data from an options file and assign it to the jobs inputdata field
    j.inputdata = getGaudiExecInputData([\"~/cmtuser/DaVinci_v22r0p2/Tutorial/Analysis/options/Bs2JpsiPhi2008.py\"], j.application)

    This is also called behind the scenes for j.readInputData([\"~/cmtuser/DaVinci_v22r0p2/Tutorial/Analysis/options/Bs2JpsiPhi2008.py\"])

    '''
    if not isinstance(optsfiles, list):
        optsfiles = [optsfiles]

    # use a dummy file to keep the parser happy
    if len(optsfiles) == 0:
        raise GangaException("Need a file to parse to call this method")

    try:
        parser = PythonOptsCmakeParser(optsfiles, app)
    except Exception as err:
        msg = 'Unable to parse the job options. Please check options files and extraopts.'
        logger.error("PythonOptsCmakeParserError:\n%s" % str(err))
        raise ApplicationConfigurationError(msg)

    return parser.get_input_data()

exportToGPI('getGaudiExecInputData', getGaudiExecInputData, 'Functions')

def prepare_cmake_app(myApp, myVer, myPath='$HOME/cmtuser', myUse=None, myFolder=None , myBranch='master'):
    """
        Short helper function for setting up minimal application environments on disk for job submission
        Args:
            myApp (str): This is the name of the app to pass to lb-dev
            myVer (str): This is the version of 'myApp' to pass to lb-dev
            myPath (str): This is where lb-dev will be run
            myUse (str): This is a git lb-use which will be run once the lb-dev has executed
            myFolder (str): This is a git lb-checkout after the lb-use. 
            myBranch (str): This is the branch used for the lb-checkout. Master branch assumed if not specified
    """

    full_path = expandfilename(myPath, True)
    if not path.exists(full_path):
        makedirs(full_path)

    #First guess a suitable platform for checking out the application
    verStat, verOut, verErr = _exec_cmd('source /cvmfs/lhcb.cern.ch/lib/LbEnv && lb-sdb-query listPlatforms %s %s' % (myApp, myVer), full_path)
    if verStat != 0 or len(verOut.decode().split('\n'))==0:
        logger.error("lb-sdb-query listPlatforms %s %s failed!" % (myApp, myVer))
        raise ApplicationPrepareError(verErr)

    platformToUse = verOut.decode().split('\n')[-2]

    if not path.exists(full_path + '/' + myApp + 'Dev_' +myVer):
        devStat, devOut, devErr = _exec_cmd('source /cvmfs/lhcb.cern.ch/lib/LbEnv && source LbLogin.sh --cmtconfig=%s && lb-dev %s/%s' % (platformToUse, myApp, myVer), full_path)
        logger.info("Running lb-dev %s %s with platform %s" % (myApp, myVer, platformToUse))
        if devStat != 0:
            logger.error("lb-dev %s %s failed!" % (myApp, myVer))
            raise ApplicationPrepareError(devErr)
    else:
        raise GangaException("Path %s already exists. Not checking out application. Try a different location." % str(full_path + '/' + myApp + 'Dev_' +myVer))
    dev_dir = path.join(full_path, myApp + 'Dev_' + myVer)
    logger.info("Set up App Env at: %s" % dev_dir)
    if myUse:
        lbUse, lbUseOut, lbUseErr = _exec_cmd('source /cvmfs/lhcb.cern.ch/lib/LbEnv && source LbLogin.sh --cmtconfig=%s && git lb-use %s' % (platformToUse, myUse), dev_dir)
        logger.info("Running git lb-use %s" % myUse)
        if lbUse != 0:
            logger.error("git lb-use %s failed!" % myUse)
            raise ApplicationPrepareError(lbUseErr)
        if myFolder:
            chk, chkOut, chkErr = _exec_cmd('source /cvmfs/lhcb.cern.ch/lib/LbEnv && source LbLogin.sh --cmtconfig=%s && git lb-checkout %s/%s %s' % (platformToUse, myUse, myBranch, myFolder), dev_dir)
            logger.info("Running git lb-checkout %s/%s %s" % (myUse,myBranch, myFolder))
            if chk != 0:
                logger.error("git lb-checkout %s/%s %s failed!" % (myUse, myBranch, myFolder))
                raise ApplicationPrepareError(chkErr)
    return dev_dir

exportToGPI('prepare_cmake_app', prepare_cmake_app, 'Functions')

def prepareGaudiExec(myApp, myVer, myPath='$HOME/cmtuser', myUse=None, myFolder=None , myBranch = 'master'):
    """
        Setup and Return a GaudiExec based upon a release version of a given App.
        Args:
            myApp (str): This is name of the App you want to run
            myVer (str): This is the version of the app you want
            myPath (str): This is where lb-dev will be run
            myUse (str): This is a git lb-use which will be run once the lb-dev has executed
            myFolder (str): This is a git lb-checkout for after the lb-use. 
            myBranch (str) : This is what is appended to the lb-checkout myFolder/myBranch , master assumed if not specified
    """
    path = prepare_cmake_app(myApp, myVer, myPath, myUse, myFolder, myBranch)
    from GangaCore.GPI import GaudiExec
    return GaudiExec(directory=path)

exportToGPI('prepareGaudiExec', prepareGaudiExec, 'Functions')

def _exec_cmd(cmd, cwdir):
    """
        This is taken from the code which runs SetupProject
        TODO: Replace me with a correct call to execute once the return code is known to work

        Args:
            cmd (str): This is the full command which is to be executed
            cwdir (str): The folder the command is to be run in
    """
    pipe = subprocess.Popen(cmd, shell=True, env={}, cwd=cwdir,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.DEVNULL)
    stdout, stderr = pipe.communicate()
    while pipe.poll() is None:
        time.sleep(0.5)
    return pipe.returncode, stdout, stderr




################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 17/01/2014
################################################################################
"""@package ND280Control
Ganga module to execute runND280 from the nd280Control package.
"""

from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *

from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.Core.exceptions import ApplicationConfigurationError

import os, shutil, re
from GangaCore.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class runND280(IPrepareApp):
    """
    runND280 application running runND280 from nd280Control.
        app = runND280()

    The required input for this module are:
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'
        app.configfile = 'mynd280config.cfg'

    NOTE: Ganga will run a bash script so the CMT setup script must be in sh/bash.

    It is also possible to pass command line arguments to the executable.
    For example to use '/home/me/tmp' as temporary directory:
        app.args = ['-t','/home/me/tmp']

    """
    _schema = Schema(Version(1,1), {
        'args' : SimpleItem(defvalue=[],typelist=['str','GangaCore.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'cmtsetup' : SimpleItem(defvalue=[],doc='Setup script(s) in bash to set up cmt and the cmt package of the executable.', typelist=['str'],sequence=1,strict_sequence=0),
        'configfile' : SimpleItem(defvalue=None,doc='Filename of the nd280Control config file.', typelist=['str','type(None)']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        } )
    _category = 'applications'
    _name = 'runND280'
    _exportmethods = ['prepare']
    _GUIPrefs = [ { 'attribute' : 'args', 'widget' : 'String_List' },
                  { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                  { 'attribute' : 'configfile', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'args', 'widget' : 'String_List' },
                          { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                          { 'attribute' : 'configfile', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]
    exe = 'runND280'

    def __init__(self):
        super(runND280,self).__init__()


    def configure(self,masterappconfig):
        
        args = convertIntToStringArgs(self.args)

        job = self.getJobObject()

        if self.cmtsetup == []:
          raise ApplicationConfigurationError('No cmt setup script given.')

        for arg in args:
          if arg == '-c':
            raise ApplicationConfigurationError('Option "-c" given in args. You must use the configfile variable instead.')

        # setup the config file for this job
        if self.configfile is None:
          raise ApplicationConfigurationError('No config file given. Use args list or configfile field.')
        # check if given config file exists
        if not os.path.exists(self.configfile):
          raise ApplicationConfigurationError('The given config file "'+self.configfile+'" was not found.')
        if not os.path.isfile(self.configfile):
          raise ApplicationConfigurationError('The given config file "'+self.configfile+'" is not a file.')

        # Right here, take the input config file and change it as needed
        # If found inputfile, just put the first file in the inputdata
        # If this is inputfile_list, then it is in cherry pick so we can put all the files from inputdata
        inConf = open(self.configfile)
        outConf = ''

        for line in inConf:
            inputfile_listfnd = re.match(r"^inputfile_list\s*=", line)
            inputfilefnd = re.match(r"^inputfile\s*=", line)
            midas_filefnd = re.match(r"^midas_file\s*=", line)
            if inputfile_listfnd or inputfilefnd or midas_filefnd:
              if job.inputdata is None:
                raise ApplicationConfigurationError('The given config file requires an input file but the inputdata of the job is not defined.')
              # TODO: Check if there is an inputdata
              infiles = job.inputdata.get_dataset_filenames()
              if len(infiles) < 1:
                raise ApplicationConfigurationError('The given config file contains "inputfile" but not input file was given')
              if inputfile_listfnd:
                line = 'inputfile_list = ' + ' '.join(infiles) + '\n'
              elif inputfilefnd:
                if len(infiles) > 1:
                  raise ApplicationConfigurationError('The given config file contains "inputfile" but more than one input file was given')
                line = 'inputfile = ' + infiles[0] + '\n'
              elif midas_filefnd:
                if len(infiles) > 1:
                  raise ApplicationConfigurationError('The given config file contains "midas_file" but more than one file was given')
                line = 'midas_file = ' + infiles[0] + '\n'
            
            outConf += line
        job.getInputWorkspace().writefile(FileBuffer('nd280Config.cfg',outConf),executable=0)

        args.append('-c')
        args.append(job.inputdir+'nd280Config.cfg')

        argsStr = ' '.join(args)
        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        for f in self.cmtsetup:
            script += 'source '+f+'\n'
        script += 'runND280 '+argsStr+'\n'
        job.getInputWorkspace().writefile(FileBuffer('runND280.sh',script),executable=1)

        self._scriptname = job.inputdir+'runND280.sh'

        return (None,None)



config = getConfig('defaults_runND280') #_Properties
# config.options['exe'].type = type(None)


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        c = StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        return c
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('runND280','LSF', RTHandler)
allHandlers.add('runND280','Local', RTHandler)
allHandlers.add('runND280','PBS', RTHandler)
allHandlers.add('runND280','SGE', RTHandler)
allHandlers.add('runND280','Slurm', RTHandler)
allHandlers.add('runND280','Condor', RTHandler)
allHandlers.add('runND280','LCG', LCGRTHandler)
allHandlers.add('runND280','gLite', gLiteRTHandler)
allHandlers.add('runND280','TestSubmitter', RTHandler)
allHandlers.add('runND280','Interactive', RTHandler)
allHandlers.add('runND280','Batch', RTHandler)
allHandlers.add('runND280','Cronus', RTHandler)
allHandlers.add('runND280','Remote', LCGRTHandler)
allHandlers.add('runND280','CREAM', LCGRTHandler)




#! /usr/bin/env python
import sys

f = file('Makefile.incl','w')

tmpl = """Makefile-%(doc)s: $(wildcard src/%(doc)s/*)
	@echo Creating Makefile-%(doc)s
	script/make-doc.py %(doc)s
include Makefile-%(doc)s
"""

for i in sys.argv[1:]:
	print >>f, tmpl % {'doc':i}
## 	print >>f,'Makefile-%s: $(wildcard ganga-%s/*)'%(i,i)
## 	print >>f,'	@echo Creating Makefile-%s'%(i)			
## 	print >>f,'	script/make-doc.py %s'%i 			
## 	print >>f,'include Makefile-%s'%i
## 	print >>f



import os
import sys
import time
import socket
import inspect
import traceback
import pickle
import uuid
from GangaCore.Runtime.GPIexport import exportToGPI
from GangaCore.GPIDev.Base.Proxy import addProxy, stripProxy
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
#from GangaCore.Core.GangaThread.WorkerThreads.WorkerThreadPool import WorkerThreadPool
#from GangaCore.Core.GangaThread.WorkerThreads.ThreadPoolQueueMonitor import ThreadPoolQueueMonitor
from GangaDirac.Lib.Utilities.DiracUtilities import execute
logger = getLogger()
#user_threadpool       = WorkerThreadPool()
#monitoring_threadpool = WorkerThreadPool()
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/#


def diracAPI(cmd, timeout=60, cred_req=None):
    '''
    Args:
        cmd (str): This is the command you want to execute from within an active DIRAC session
        timeout (int): This is the maximum time(sec) the session has to complete the task
        cred_req (ICredentialRequirement): This is the (optional) credential passed to construct the correct DIRAC env

    Execute DIRAC API commands from w/in GangaCore.

    The stdout will be returned, e.g.:

    # this will simply return 87
    diracAPI(\'print 87\')

    # this will return the status of job 66
    # note a Dirac() object is already provided set up as \'dirac\'
    diracAPI(\'print(Dirac().getJobStatus([66]))\')
    diracAPI(\'print(dirac.getJobStatus([66]))\')

    # or can achieve the same using command defined and included from
    # getConfig('DIRAC')['DiracCommandFiles']
    diracAPI(\'status([66])\')

    '''
    return execute(cmd, timeout=timeout, cred_req=cred_req)

exportToGPI('diracAPI', diracAPI, 'Functions')

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/#

running_dirac_process = False
dirac_process = None
dirac_process_ids = None
def startDiracProcess():
    '''
    Start a subprocess that runs the DIRAC commands
    '''
    HOST = 'localhost'  #Connect to localhost
    end_trans = '###END-TRANS###'
    import subprocess
    from GangaDirac.Lib.Utilities.DiracUtilities import getDiracEnv, getDiracCommandIncludes, GangaDiracError
    global dirac_process
    #Some magic to locate the python script to run
    from GangaDirac.Lib.Server.InspectionClient import runClient
    #Create a socket and bind it to 0 to find a free port
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind((HOST, 0))
    PORT = s.getsockname()[1]
    s.close()
    #Pass the port no as an argument to the popen
    serverpath = os.path.join(os.path.dirname(inspect.getsourcefile(runClient)), 'DiracProcess.py')
    popen_cmd = ['python',serverpath, str(PORT)]
    dirac_process = subprocess.Popen(popen_cmd, env = getDiracEnv(), stdin=subprocess.PIPE)
    global running_dirac_process
    running_dirac_process = (dirac_process.pid, PORT)

    #Now set a random string to make sure only commands from this sessions are executed
    rand_hash = uuid.uuid4()
    global dirac_process_ids
    dirac_process_ids = (dirac_process.pid, PORT, rand_hash)
    #Pipe the random string without waiting for the process to finish.
    dirac_process.stdin.write(str(rand_hash).encode("utf-8"))
    dirac_process.stdin.close()

    data = ''
    #We have to wait a little bit for the subprocess to start the server so we try until the connection stops being refused. Set a limit of one minute.
    connection_timeout = time.time() + 60
    started = False
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while time.time()<connection_timeout and not started:
        try:
            s.connect((HOST, PORT))
            started = True
        except socket.error as serr:
            time.sleep(1)
    if not started:
        raise GangaDiracError("Failed to start the Dirac server process!")
    #Now setup the Dirac environment in the subprocess
    dirac_command = str(rand_hash)
    dirac_command = dirac_command + getDiracCommandIncludes()
    dirac_command = dirac_command + end_trans
    s.sendall(dirac_command.encode("utf-8"))
    data = s.recv(1024)
    s.close()

exportToGPI('startDiracProcess', startDiracProcess, 'Functions')

def stopDiracProcess():
    '''
    Stop the Dirac process if it is running
    '''
    global running_dirac_process
    if running_dirac_process:
        logger.info('Stopping the DIRAC process')
        dirac_process.kill()
        running_dirac_process = False

exportToGPI('stopDiracProcess', stopDiracProcess, 'Functions')

def diracAPI_interactive(connection_attempts=5):
    '''
    Run an interactive server within the DIRAC environment.
    '''

    from GangaDirac.Lib.Server.InspectionClient import runClient
    serverpath = os.path.join(os.path.dirname(inspect.getsourcefile(runClient)), 'InspectionServer.py')
    from GangaCore.Core.GangaThread.WorkerThreads import getQueues
    getQueues().add(execute("execfile('%s')" % serverpath, timeout=None, shell=False))

    #time.sleep(1)
    sys.stdout.write( "\nType 'q' or 'Q' or 'exit' or 'exit()' to quit but NOT ctrl-D")
    i = 0
    excpt = None
    while i < connection_attempts:
        try:
            runClient()
            break
        except:
            if i == (connection_attempts - 1):
                excpt = traceback.format_exc()
        finally:
            i += 1
    return excpt
exportToGPI('diracAPI_interactive', diracAPI_interactive, 'Functions')

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/#


def diracAPI_async(cmd, timeout=120):
    '''
    Execute DIRAC API commands from w/in GangaCore.
    '''
    from GangaCore.Core.GangaThread.WorkerThreads import getQueues
    return getQueues().add(execute(cmd, timeout=timeout))

exportToGPI('diracAPI_async', diracAPI_async, 'Functions')

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/#


def getDiracFiles():
    from GangaDirac.Lib.Files.DiracFile import DiracFile
    from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
    filename = DiracFile.diracLFNBase().replace('/', '-') + '.lfns'
    logger.info('Creating list, this can take a while if you have a large number of SE files, please wait...')
    execute('dirac-dms-user-lfns &> /dev/null', shell=True, timeout=None)
    g = GangaList()
    with open(filename[1:], 'r') as lfnlist:
        lfnlist.seek(0)
        g.extend((DiracFile(lfn='%s' % lfn.strip()) for lfn in lfnlist.readlines()))
    return addProxy(g)

exportToGPI('getDiracFiles', getDiracFiles, 'Functions')

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/#


def dumpObject(object, filename):
    '''
    These are complimentary functions to export/load which are already exported to
    the GPI from GangaCore.GPIDev.Persistency. The difference being that these functions will
    export the objects using the pickle persistency format rather than a Ganga streaming
    (human readable) format.
    '''
    try:
        with open(os.path.expandvars(os.path.expanduser(filename)), 'wb') as f:
            pickle.dump(stripProxy(object), f)
    except:
        logger.error("Problem when dumping file '%s': %s" % (filename, traceback.format_exc()))
exportToGPI('dumpObject', dumpObject, 'Functions')


def loadObject(filename):
    '''
    These are complimentary functions to export/load which are already exported to
    the GPI from GangaCore.GPIDev.Persistency. The difference being that these functions will
    export the objects using the pickle persistency format rather than a Ganga streaming
    (human readable) format.
    '''
    try:
        with open(os.path.expandvars(os.path.expanduser(filename)), 'rb') as f:
            r = pickle.load(f)
    except:
        logger.error("Problem when loading file '%s': %s" % (filename, traceback.format_exc()))
    else:
        return addProxy(r)
exportToGPI('loadObject', loadObject, 'Functions')

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/#
